/****************************************************************************
 *
 *   File: finch_interface.cc:
 *   
 *   Author: Mark Stephenson
 *   Date: Tue Jun  3 17:27:15 2003
 *
 *   Function: The interface to finch's library.
 *
 ***************************************************************************/


#include "finch_interface.h"
#include "configure_expr.h"
#include <stdio.h>
#include <stdarg.h>

static genetic_expr     *FINCH_curr_expr = NULL;
static genetic_expr_bnf *FINCH_curr_bnf  = NULL;
static double            FINCH_srate     = 0.0;

void
FINCH_initialize_lib ()
///////////////////////////////////////////////////////////////
//
// Initialize the finch library.  This function should
// only be called once-- when your pass initializes.
// 
///////////////////////////////////////////////////////////////
{
  // Get the FINCH_HOME environment variable.  This 
  // is the directory that is used to host the config file.
  // The directory $FINCH_HOME/evaluate is used for 
  // communication, such as the priority functions, fitnesses
  // etc.
  char* finch_home = getenv ("FINCH_HOME");

  // The FINCH_EVAL_DIR is used by the pbs_harness.
  char* finch_eval_dir = getenv ("FINCH_EVAL_DIR");

  finch_claim ((finch_home != NULL), 
	       "Please set the FINCH_HOME environment variable.");
  finch_claim ((FINCH_curr_expr == NULL),
	       "Please only call FINCH_initialize_lib once.");
  
  // If the config file is not present, we have to create it.
  string config_fname = string(finch_home) + string("/config");
  string eval_dir = finch_eval_dir ? 
    string(finch_eval_dir) : string(finch_home);
  eval_dir += string("/evaluate/");
  string pfun_fname = eval_dir + string("pfun.l");

  // Otherwise, go ahead and create the genetic expressions.
  ifstream config_file (config_fname.c_str());
  FINCH_curr_bnf = (config_file) ? 
    new genetic_expr_bnf_default (config_file) : NULL;
    
  if (FINCH_curr_bnf)
  {
    FINCH_curr_expr = genetic_expr::parse_expr (FINCH_curr_bnf, 
						pfun_fname.c_str());
    cout << "Using priority function in: " << pfun_fname << endl;
    finch_claim (FINCH_curr_expr, 
		 "No priority function in $FINCH_HOME/evaluate/pfun.l.");
  }
  else
  {
    FINCH_curr_expr = new configure_expr (config_fname);
  }

  string srate_fname = string(finch_home) + string("/srate");
  ifstream srate(srate_fname.c_str());
  if (srate) srate >> FINCH_srate;
  finch_claim (FINCH_srate <= 1.0 && FINCH_srate >= 0.0,
	       "Invalid sample rate.");
}


///////////////////////////////////////////////////////////////
//
// Evaluate the current priority function with the arguments
// that are passed in.
//
///////////////////////////////////////////////////////////////
double 
FINCH_evaluate_real_expr (unsigned num_dbls, unsigned num_bools, ...)
{
  va_list ap;
  double *dlist;
  int32  *blist;

  dlist = new double[num_dbls];
  blist = new int32[num_bools];

  va_start (ap, num_bools);
  for (uint32 i = 0; i < num_dbls; i++)
    dlist[i] = va_arg(ap, double);
  for (uint32 j = 0; j < num_bools; j++)
    blist[j] = va_arg(ap, int32);
  va_end (ap);

  lval val = FINCH_curr_expr->evaluate (dlist, blist, 
					num_dbls, num_bools,
					FINCH_srate);

  delete[] dlist;
  delete[] blist;
  return val.d;
}

int 
FINCH_evaluate_binary_expr (unsigned num_dbls, unsigned num_bools, ...)
{
  va_list ap;
  double *dlist;
  int32  *blist;

  dlist = new double[num_dbls];
  blist = new int32[num_bools];

  va_start (ap, num_bools);
  for (uint32 i = 0; i < num_dbls; i++)
    dlist[i] = va_arg(ap, double);
  for (uint32 j = 0; j < num_bools; j++)
    blist[j] = va_arg(ap, int32);
  va_end (ap);

  lval val = FINCH_curr_expr->evaluate (dlist, blist, 
					num_dbls, num_bools,
					FINCH_srate);

  delete[] dlist;
  delete[] blist;
  return val.i;
}

double 
FINCH_evaluate_real_array (unsigned num_dbls, unsigned num_bools,
                           double *dbls, int *bools)
{
  lval val = FINCH_curr_expr->evaluate (dbls, bools, 
					num_dbls, num_bools,
					FINCH_srate);
  return val.d;
}

int
FINCH_evaluate_binary_array (unsigned num_dbls, unsigned num_bools,
                             double *dbls, int *bools)
{
  lval val = FINCH_curr_expr->evaluate (dbls, bools, 
					num_dbls, num_bools,
					FINCH_srate);
  return val.i;
}

///////////////////////////////////////////////////////////////
//
// Delete any dynamically allocated memory that we have
// created.
//
///////////////////////////////////////////////////////////////
void   
FINCH_finalize_lib ()
{
  char* finch_eval_dir = getenv ("FINCH_HOME");
  string eval_dir = string(finch_eval_dir) + string("/notes/");
  char *tname = tempnam(eval_dir.c_str(), "_note");

  ofstream annotated (tname);
  if (annotated) {
    FINCH_curr_expr->print (annotated, true);
    annotated.close ();
  }

  delete FINCH_curr_expr;
  delete FINCH_curr_bnf;
}


void   
FINCH_dump_features (char *filename, 
		     unsigned num_dbls, unsigned num_bools,
		     double *dbls, int *bools)
///////////////////////////////////////////////////////////////
//
// Dumps features to the specified file.
//
///////////////////////////////////////////////////////////////
{
  FILE *dumpfile = fopen (filename, "a");
  assert (dumpfile);
  for (uint32 i = 0; i < num_dbls; i++)
    fprintf (dumpfile, "%f ", dbls[i]);
  for (uint32 j = 0; j < num_bools; j++)
    fprintf (dumpfile, "%d ", bools[j]);
  fprintf (dumpfile, "\n");
  fclose (dumpfile);
}
