/******************************************************************
 *
 *   File: genetic_expr.cc:
 *   
 *   Author: Mark Stephenson
 *   Date: Mon Jul 22 14:40:00 2002
 *
 *   Function: BNF implementation of the language.
 *
 *****************************************************************/

#include "genetic_expr.h"
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <list>

// Represents unitialized fitnesses.
#define NOT_INITIALIZED -1.0


///////////////////////////////////////////
////// genetic_expr class members. ////////
///////////////////////////////////////////

genetic_expr*
genetic_expr::random (genetic_expr_bnf *bnf,
		      uint32 h, 
		      genetic_expr* parent,  /* NULL */
		      uint32 type)           /* DOUBLE */
/////////////////////////////////////////////
//  Returns a randomly generated expression.
//  This function uses the dispatch tables
//  in global.* to create new expressions.
/////////////////////////////////////////////
{
  expr_nonterm_fun* nonterms;
  expr_term_fun* terms;
  uint32 num_nonterms;
  uint32 num_terms;
  int32 which;

  switch (type) {
  case DOUBLE:
    nonterms = bnf->dbl_nonterms ();
    num_nonterms = bnf->num_dbl_nonterms ();
    terms = bnf->dbl_terms ();
    num_terms = bnf->num_dbl_terms ();
    break;
  case BOOLEAN:
    nonterms = bnf->bool_nonterms ();
    num_nonterms = bnf->num_bool_nonterms ();
    terms = bnf->bool_terms ();
    num_terms = bnf->num_bool_terms ();
    break;
  default:
    assert (0);
    break;
  }
 
  // If the height has reached zero,
  // then return a terminal.
  if (h == 0) {
    which = ::random() % num_terms;
    return terms[which](bnf, parent);
  }

  which = ::random() % num_nonterms;
  return nonterms[which](bnf, h-1, parent);
}

genetic_expr*
genetic_expr::parse_expr (genetic_expr_bnf* bnf, istream& stm) 
/////////////////////////////////////////////
// Creates the expression described in the
// file.
/////////////////////////////////////////////
{
  genetic_lexer lexer(&stm);
  return parse_expr (bnf, NULL, &lexer);
}

genetic_expr*
genetic_expr::parse_expr (genetic_expr_bnf* bnf, const char *expr_fname) 
/////////////////////////////////////////////
// Creates the expression described in the
// file.
/////////////////////////////////////////////
{
  ifstream stm(expr_fname);
  if (stm) 
  {
    genetic_lexer lexer(&stm);
    return parse_expr (bnf, NULL, &lexer);
  }
  return NULL;
}

void
genetic_expr::mutate ()
///////////////////////////////////////////
// Finds a random subexpression and 
// mutates it.
///////////////////////////////////////////
{
  genetic_expr *e = this;
  double w = 0.0;
  bool internal_node;

  // Choose an internal node 50% of the time.
  internal_node = (drand48() < 0.50);
  choose_random_node (internal_node, &w, &e);
  assert (e);
  e->mutate_node ();
}

void
genetic_expr::mutate_node () 
///////////////////////////////////////////
// Mutates this node by killing a random 
// child and replacing it with a random
// expression.
///////////////////////////////////////////
{
  genetic_expr *mutate_node;
  uint32 mutate_point;
  uint32 height;
  uint32 type;

  // Override virtual function for expressions
  // that have no children.
  if (num_rands() == 0) return; 
  mutate_point = ::random() % num_rands();
  mutate_node = rands()[mutate_point];

  // Make the height the same as the original 
  // subtree.
  height = mutate_node->height ();
  type = mutate_node->typing ();
  if (height > 5) height = 5;
  rands()[mutate_point] = genetic_expr::random (m_bnf, height, this, type);

  delete mutate_node;
}

genetic_expr*
genetic_expr::hvl_mutate ()
///////////////////////////////////////////
// Hierarchical variable length mutate,
// as described in Una-May's thesis. This
// function returns a mutated copy of 
// this expression.
///////////////////////////////////////////
{
  uint32 hvl_muttype;
  uint32 type;
  uint32 hite;
  uint32 maxi;
  uint32 i;
  genetic_expr *e;
  genetic_expr *tmp;
  genetic_expr *par;
  genetic_expr *rnd;
  genetic_expr *maxc;
  genetic_expr *child;

  // Choose a random node.
  tmp = this->clone ();
  e = tmp->random_subexpr ();
  assert (e);

  type = e->typing ();
  par = e->get_parent ();
  hite = e->height();

  hvl_muttype = ::random() % 3;
  cout << "TYPE: " << hvl_muttype << "\nEXP: ";
  this->print (cout);
  cout << "\nSEL: ";
  e->print (cout);
  
  switch (hvl_muttype)
  {
  case 0:
    // Substitution: Create a new random
    // node of same basic arity and type.
    rnd = random (m_bnf, hite, par, type);
    if (e != tmp) 
      par->supplant (e, rnd);
    else 
      tmp = rnd;
    delete e;
    break;

  case 1:
    // Insertion: Make a new node with
    // the chosen node as the first
    // child, create new leaves.
    rnd = random (m_bnf, hite, par, type);
    cout << "\nRND: ";
    rnd->print (cout);
    if (e != tmp)
      par->supplant (e, rnd);
    else
      tmp = rnd;

    // Find first child with same type.
    child = NULL;
    for (i = 0; i < rnd->num_rands(); i++) 
    {
      child = rnd->rands()[i];
      if (child->typing() == type)
	break;
    }

    if (i >= rnd->num_rands() || child == NULL)
      delete e;
    else
    {
      // Found one. Remove and replace 
      // the match.
      e->set_parent (rnd);
      rnd->supplant (child, e);
      delete child;
    }
    break;

  case 2:
  default:
    // Deletion: Delete node and move 
    // largest child of same type up.
    maxc = NULL;
    maxi = 0;

    for (i = 0; i < e->num_rands(); i++) 
    {
      child = e->rands()[i];
      if ((child->size() >= maxi) && 
	  (child->typing() == type)) {
        maxi = (int) child->size();
	maxc = child;
      }
    }

    if (maxc == NULL)
      // Couldn't find an appropriately 
      // typed subexpression, so create
      // a random one.
      rnd = random (m_bnf, hite, par, type);
    else
      // Replace the node with the 
      // biggest child we found.
      rnd = maxc->clone ();
    
    rnd->set_parent (par);

    if (e != tmp) 
      par->supplant (e, rnd);
    else
      tmp = rnd;

    delete e;
    break;
  }

  cout << endl << "MUT: ";
  tmp->print (cout);
  cout << endl;
  cout << endl;

  return tmp;
}


lval
genetic_expr::evaluate (double dbls[], 
			int32 bools[], 
			uint32 num_dbls,
			uint32 num_bools,
			double sample_rate)
/////////////////////////////////////////////
// The only class that does anything with
// num_dbls and num_bools is configure_expr
// so just call the normal evaluate.
/////////////////////////////////////////////
{
  bool sample = (drand48() < sample_rate);
  return this->evaluate (dbls, bools, sample);
}


void
genetic_expr::cross (genetic_expr *e, 
		     genetic_expr*& child1,
		     genetic_expr*& child2)
/////////////////////////////////////////////
// Crosses this expression with e. The two 
// children are returned as call by 
// reference parameters.
/////////////////////////////////////////////
{
  genetic_expr *st1;
  genetic_expr *st2;
  genetic_expr *st1_parent;
  genetic_expr *st2_parent;
  genetic_expr *rand1;
  genetic_expr *rand2;
  uint32 not_same_type;
  uint32 st1_height;
  uint32 st2_height;

  child1 = e->clone ();
  child2 = this->clone ();

  // Find two subtrees.
  not_same_type = find_subtrees (child1, child2, st1, st2);
  
  // Find the parents of the subtrees.
  st1_parent = st1->get_parent ();
  st2_parent = st2->get_parent ();

  if (not_same_type) {
    // The types are not the same, so we can't legally 
    // swap the two subtrees.  Instead, we'll create
    // two random subtrees that are about the same
    // height as the amputated subtrees. But don't let
    // the trees grow too big.
    st1_height = st1->height ();
    st2_height = st2->height ();
    if (st1_height > 5) st1_height = 5;
    if (st2_height > 5) st2_height = 5;

    rand1 = random (m_bnf, st2_height, st1_parent, st1->typing());
    rand2 = random (m_bnf, st1_height, st2_parent, st2->typing());

    if (st1_parent != NULL) 
      st1_parent->supplant (st1, rand1);
    else 
      // We selected the root node.
      child1 = rand1;
    
    if (st2_parent != NULL) 
      st2_parent->supplant (st2, rand2);
    else
      child2 = rand2;

    delete st1;
    delete st2;
  } else {
    // The subtrees are of the same type, so we can
    // simply swap them.
    st2->set_parent (st1_parent);
    st1->set_parent (st2_parent);
    
    if (st1_parent != NULL) 
      st1_parent->supplant (st1, st2);
    else
      // We selected the root node.
      child1 = st2;

    if (st2_parent != NULL) 
      st2_parent->supplant (st2, st1);
    else
      child2 = st1;
  }
}

bool
genetic_expr::find_subtrees (genetic_expr *e1,
			     genetic_expr *e2,
			     genetic_expr *&st1,
			     genetic_expr *&st2) const
////////////////////////////////////////////
// Finds two subtrees to cross.
////////////////////////////////////////////
{
  uint32 st1_type;
  uint32 st2_type;
  bool not_same_type = true;
  
  for (uint32 i = 0; not_same_type && i < 1; i++) {
    st2 = e2->random_subexpr ();
    st2_type = st2->typing ();
    for (uint32 j = 0; not_same_type && j < 2; j++) {
      st1 = e1->random_subexpr ();
      st1_type = st1->typing ();
      not_same_type = (st1_type != st2_type);
    }
  }

  return not_same_type;
}


void
genetic_expr::record (lval v, bool sample)
////////////////////////////////////////////
//  Profiles the expression by recording
//  the values that it takes on.
////////////////////////////////////////////
{
  if (sample) 
  {
    if (!m_pdata) m_pdata = new profile_data;
    m_pdata->data.push_back(v);
  }
}


void
genetic_expr::merge (genetic_expr *e, bool concat)
////////////////////////////////////////////
//  Merges the profile data.
////////////////////////////////////////////
{
  assert (concat && e);
  if (!(this->m_pdata)) m_pdata = new profile_data;

  if (e->m_pdata)
  {
    // Push all of e's values on to this's list.
    list<lval>::iterator first = e->m_pdata->data.begin ();
    list<lval>::iterator last = e->m_pdata->data.end ();
    while (first != last)
      this->m_pdata->data.push_back(*first++);
  }
   
  for (unsigned i = 0; i < this->num_rands(); i++) 
  {
    cout << "Merging for " << e->opcode() << endl;
    assert (e->opcode() == this->opcode());
    ((this->rands())[i])->merge ((e->rands())[i], concat);
  }
}


double
genetic_expr::correlation (const genetic_expr *e) const
////////////////////////////////////////////
//  Computes the correlation of this expr
//  with the one that's passed in.
//  
//  std_mean must be called on this and e
//  before this function is called.
////////////////////////////////////////////
{
  double mean_x;
  double mean_y;
  double stdx;
  double stdy;
  double num;
  double den;
  double cor;
  double mean_xy = 0.0;
  double N;

  profile_data* x = e->m_pdata;

  assert (x);
  assert (m_pdata);

  N = m_pdata->data.size();
  assert (x->data.size() == N);

  stdx = x->stddev;
  stdy = m_pdata->stddev;
  mean_x = x->mean;
  mean_y = m_pdata->mean;

  list<lval>::iterator first = m_pdata->data.begin ();
  list<lval>::iterator last = m_pdata->data.end ();
  list<lval>::iterator firste = x->data.begin ();
  while (first != last) 
  {
    mean_xy += ((this->todouble(*first))*(e->todouble(*firste)));
    first++;
    firste++;
  }
  mean_xy = mean_xy / N;
  num = mean_xy - mean_x * mean_y;
  den = stdx * stdy;
  assert (!isnan(den) && !isnan(num));
  cor = (den == 0.0) ? 0.0 : (num / den);
  return cor;
}


void
genetic_expr::std_mean () const
////////////////////////////////////////////
//  Computes the stddev and mean of 
//  this expr.
////////////////////////////////////////////
{
  assert (m_pdata && (m_pdata->data.size() > 0));

  double mean = 0.0;
  double std = 0.0;
  double val;
  double N = m_pdata->data.size();
  
  list<lval>::iterator first = m_pdata->data.begin ();
  list<lval>::iterator last = m_pdata->data.end ();
  while (first != last) 
    mean += todouble(*first++);
  mean = mean / N;

  first = m_pdata->data.begin ();
  while (first != last) 
  {
    val = todouble(*first);
    std += (val * val);
    first++;
  }
  std = std / N;
  std -= (mean * mean);
  m_pdata->mean = mean;
  m_pdata->stddev = (std < 0.0001) ? 0.0 : sqrt(std);
}


void 
genetic_expr::set_fitness_element (uint32 id, double fitness)
////////////////////////////////////////////
//  Sets the fitness of the expression
//  for a given id.
////////////////////////////////////////////
{
  uint32 size = m_fitness_vector.size();
  if (id >= size) {
    m_fitness_vector.resize (id+1);
    for (uint32 i = size; i < id; i++)
      m_fitness_vector[i] = NOT_INITIALIZED;
  }

  m_fitness_vector[id] = fitness;
}

double 
genetic_expr::get_fitness_element (uint32 id) const
////////////////////////////////////////////
//  Gets the fitness for a given id.
////////////////////////////////////////////
{
  assert (id < m_fitness_vector.size());
  return m_fitness_vector[id];
}

bool 
genetic_expr::memoized (uint32 id) const
////////////////////////////////////////////
//  Return true if this expression has
//  been memoized for a given id.
////////////////////////////////////////////
{
  if (id >= m_fitness_vector.size()) return false;
  return (m_fitness_vector[id] != NOT_INITIALIZED);
}

uint32
genetic_expr::size () const
////////////////////////////////////////////
//  Returns the number of nodes in this
//  expression.
////////////////////////////////////////////
{
  uint32 s = 1;
  for (uint32 i = 0; i < this->num_rands(); i++) 
    s += rands()[i]->size ();

  return s;
}

void
genetic_expr::print (ostream& os, bool state) const
/////////////////////////////////////////////
// Prints the expression to the stream.
/////////////////////////////////////////////
{
  print_prologue (os, state);
  os << this->opcode();
  for (uint32 i = 0; i < this->num_rands(); i++) {
    rands()[i]->print (os, state);
  }
  print_epilogue (os, state);
}

void
genetic_expr::stat_print (ostream& os) const
/////////////////////////////////////////////
// Prints the expression, as well as 
// annotated statistics to a file that
// can be compiled by dot.
/////////////////////////////////////////////
{
  unsigned gid = 0;
  unsigned cid = 0;

  this->std_mean ();
  os << "digraph expr {\n";
  os << "\tordering=out;\n\tcenter=1;\n\trotate=90;\n";
  os << "\tsize=\"8.5,11\";\n";
  os << "\tnode [shape=box,fontcolor=grey,fillcolor=\"#000000\"";
  os << ",style=filled] node0;\n";
  os << "\tnode0 [label=\"" << this->opcode() << "\"];\n";
  for (uint32 i = 0; i < this->num_rands(); i++) {
    cid = this->rands()[i]->stat_print (os, this, gid);
    os << "\tnode0 -> node" << cid << ";\n";
  }  
  os << "\n}\n";
}

uint32
genetic_expr::get_corr_color (double c) const
/////////////////////////////////////////////
//
//  Returns a gray-scale RGB color based on
//  the correlation.  The higher the corr.
//  the darker the color.
//
/////////////////////////////////////////////
{
  double scale;
  uint32 tmp;
  uint32 color;

  scale = (1.0 - fabs(c)) * (double)0xff;
  tmp = (uint32)scale;
  color = (tmp << 16) | (tmp << 8) | tmp;

  return color;
}

uint32
genetic_expr::stat_print (ostream& os, 
			  const genetic_expr* p,
			  uint32& gid) const
/////////////////////////////////////////////
// Prints out the expression in dot format,
// annotated with statistics information
// (if the expr has been profiled).
//
// std_mean() must have already been called
// on p.
/////////////////////////////////////////////
{
  double corr;
  uint32 color;
  uint32 myid;
  uint32 cid;

  myid = ++gid;

  this->std_mean ();
  corr = this->correlation (p);
  color = get_corr_color (corr);
 
  os << "\tnode [shape=box,fontcolor=gray,fillcolor=\"#";
  os.precision(8);
  os << ::hex << color;
  os << "\",style=filled] node" << myid << ";\n";
  os << "\tnode" << myid << " [label=\"";

  if (this->is_leaf()) 
    this->print(os);
  else
    os << this->opcode();

  os.precision(2);
  os << " " << m_pdata->stddev << "\"];\n";  
  for (uint32 i = 0; i < this->num_rands(); i++) {
    cid = this->rands()[i]->stat_print (os, p, gid);
    os << "\tnode" << myid << " -> node" << cid << ";\n";    
  } 

  return myid;
}

uint32
genetic_expr::height () const
/////////////////////////////////////////////
// Returns the height of this expression.
/////////////////////////////////////////////
{
  uint32 tmp;
  uint32 max = 0;

  for (uint32 i = 0; i < this->num_rands(); i++) {
    tmp = rands()[i]->height ();
    if (tmp > max)
      max = tmp;
  }

  return max + 1;
}

void
genetic_expr::supplant (genetic_expr* old_expr,
			genetic_expr *new_expr)
/////////////////////////////////////////////
// Replaces one expression* with another
// expression* in this expression's list
// of operands.
/////////////////////////////////////////////
{
  for (uint32 i = 0; i < this->num_rands(); i++) {
    if (rands()[i] == old_expr)
      rands()[i] = new_expr;
  }
}

void
genetic_expr::clone_rands (genetic_expr* that)
/////////////////////////////////////////////
// Copies the rands in *that to *this.
/////////////////////////////////////////////
{
  genetic_expr *tmp;

  for (uint32 i = 0; i < that->num_rands(); i++) {
    tmp = that->rands()[i]->clone ();
    tmp->set_parent (this);
    rands()[i] = tmp;
  }
}

genetic_expr*
genetic_expr::random_subexpr ()
/////////////////////////////////////////////
// Returns a random subexpression following
// the Koza 90/10 rule: 90% of the time 
// we'll return an internal node.
/////////////////////////////////////////////
{
  double maxw;
  genetic_expr *e;

  maxw = 0.0;
  e = this;
  choose_random_node ((drand48() < 0.90), &maxw, &e);
  assert (e);
  return e;
}

void
genetic_expr::choose_random_node (bool select_internal,
				  double* max_weight,
				  genetic_expr** expr)
/////////////////////////////////////////////
// random_subexpr helper function.
/////////////////////////////////////////////
{
  double w;

  if ((is_leaf() && !select_internal) ||
      (!is_leaf() && select_internal)) 
  {
    w = drand48 ();
    if (w > *max_weight) {
      *max_weight = w;
      *expr = this;
    }      
  }

  for (uint32 i = 0; i < num_rands(); i++) {
    rands()[i]->choose_random_node (select_internal, 
				    max_weight, expr);
  }
}

void
genetic_expr::print_prologue (ostream& os, 
			      bool state = false) const
//////////////////////////////////////////////
// Prints the beginning of the expression, 
// and some state if the bool is true.
//////////////////////////////////////////////
{
  if (state && !m_parent)
  {
    // Print out the fitness vector.
    os << "[ ";
    os.setf (ios::fixed, ios::floatfield);
    for (unsigned i = 0; i < m_fitness_vector.size(); i++)
    {
      double tmp = m_fitness_vector[i];
      if (tmp != -1)
	os << i << ":" << tmp << " ";
    }
    os << "]\n";
  }
  os << " (";
  if (state)
    print_pdata (os);
}

void
genetic_expr::print_epilogue (ostream& os, bool state = false) const
//////////////////////////////////////////////
// Prints the end of the expression, 
// and some state if the bool is true.
//////////////////////////////////////////////
{
  os << ")";
}

void
genetic_expr::print_pdata (ostream& os) const
////////////////////////////////////////////
//  Prints the profile-extracted data.
////////////////////////////////////////////
{
  uint32 old_precision;

  if (m_pdata)
  {
    list<lval>::iterator first = m_pdata->data.begin ();
    list<lval>::iterator last = m_pdata->data.end ();
    os << "{";
    switch (this->typing())
    {
    case DOUBLE:
      old_precision = os.precision ();
      os.setf (ios::fixed, ios::floatfield);
      os.precision (2);

      while (first != last)
	os << (*first++).d << " ";

      os.precision (old_precision);
      break;
    case BOOLEAN:
      while (first != last)
	os << (*first++).i << " ";
      break;
    default:
      assert (0);
    }
    os << "}";
  }
}

bool 
genetic_expr::operator == (const genetic_expr& e) const
/////////////////////////////////////////////
// Returns true iff the expressions are 
// equal.
/////////////////////////////////////////////
{
  return false;
}

genetic_expr*
genetic_expr::parse_bool_expr (genetic_expr_bnf *bnf,
			       genetic_expr* parent,
			       genetic_lexer* lexer)
/////////////////////////////////////////////
// Calls the parse_expr function and makes 
// sure that it returns a boolean expression.
/////////////////////////////////////////////
{
  genetic_expr *tmp;
  
  tmp = parse_expr (bnf, parent, lexer);
  if (tmp->typing() != BOOLEAN)
    lexer->error ("Expecting a boolean value.",
		  tmp->opcode());
  return tmp;
}

genetic_expr*
genetic_expr::parse_dbl_expr (genetic_expr_bnf *bnf,
			      genetic_expr* parent,
			      genetic_lexer* lexer)
/////////////////////////////////////////////
// Calls the parse_expr function, and makes
// sure that it returns a double expression.
/////////////////////////////////////////////
{
  genetic_expr *tmp;
  
  tmp = parse_expr (bnf, parent, lexer);
  if (tmp->typing() != DOUBLE)
    lexer->error ("Expecting a double precision value.\n");
  return tmp;
}

profile_data*
genetic_expr::parse_pdata (genetic_lexer* lexer)
//////////////////////////////////////////////
// Parses the pdata tags of the expression.
// These tags are of the form:
// <n1 n2 ... nN>
//////////////////////////////////////////////
{
  uint32 tok;
  lval val;
  profile_data* tmp;

  tmp = new profile_data;
  do {
    tok = lexer->token();
    switch (tok)
    {
    case DOUBLE:
      val.d = lexer->double_val ();
      tmp->data.push_back (val);
      break;
    case BOOLEAN:
      val.i = lexer->int_val ();
      tmp->data.push_back (val);
      break;
    case RCURLY:
      break;
    default:
      finch_claim (0, "Expecting a value or a '}'.");
    }
  } while (tok != RCURLY);
  
  return tmp;
}

genetic_expr*
genetic_expr::parse_expr (genetic_expr_bnf *bnf,
			  genetic_expr* parent,
			  genetic_lexer* lexer)
/////////////////////////////////////////////
// A simple parser that uses the dispatch
// tables in global.* to call the correct
// parse functions.
/////////////////////////////////////////////
{
  genetic_expr *ret_val;
  expr_parse_fun fun;
  profile_data *pdata;
  uint32 tok;

  switch (lexer->token()) {
  case LBRACKET:
  {
    // Get the memoized values.
    list< pair<int,double> > memos;
    tok = lexer->token ();
    while (tok == INTVAL)
    {
      pair<int,double> p;
      p.first = lexer->int_val();
      if (lexer->token() != COLON)
	lexer->error ("Expecting a `:`");
      if (lexer->token() != DBLVAL)
	lexer->error ("Expecting a fitness (double)");
      p.second = lexer->double_val();
      memos.push_back (p);
      tok = lexer->token ();
    }
    // Here ends the memoized values.
    if (tok != RBRACKET) {
      lexer->error ("Expecting a ']'");
    }

    // Get the expression now.
    ret_val = parse_expr (bnf, parent, lexer);

    // Assign the memoized values to the expression.
    list< pair<int,double> >::iterator first = memos.begin ();
    list< pair<int,double> >::iterator last = memos.end ();
    while (first != last) 
    {
      ret_val->set_fitness_element ((*first).first, 
				    (*first).second);
      first++;
    }

    break;
  }
  case RBRACKET:
    lexer->error ("Unexpected ']'");
    break;
  case LPAREN:
    pdata = NULL;
    tok = lexer->token ();
    if (tok == LCURLY) {
      pdata = parse_pdata (lexer);
      tok = lexer->token ();
    }
    if (tok != FUNCTION)
      lexer->error ("Expecting a function.");

    // Find the parse rule associated with the function.
    fun = bnf->parser_binding (lexer->function_name());
    if (fun)
      ret_val = fun (bnf, parent, lexer);
    else 
      lexer->error ("Unknown function: ", 
		    lexer->function_name());

    if (lexer->token() != RPAREN)
      lexer->error ("Expecting ')'");

    ret_val->set_pdata(pdata);
    break;
  case RPAREN:
    lexer->error ("Unexpected ')'");
    break;
  case FILEEND:
  case AT:
    ret_val = NULL;
    break;
  default:
    lexer->error ("Parser error");
    break;
  }

  return ret_val;
}

genetic_expr::genetic_expr (genetic_expr_bnf *bnf,
			    genetic_expr* parent,
			    uint32 num_rands) :
/////////////////////////////////////////////
// Creates space for operands.
/////////////////////////////////////////////
  m_bnf (bnf),
  m_parent(parent),
  m_num_rands(num_rands),
  m_lifetime (0),
  m_annotation (0),
  m_expiry (0),
  m_total_fitness (0.0),
  m_fitness_vector(18),
  m_pdata (NULL)
{
  m_rands = new genetic_expr*[num_rands];
  for (uint32 i = 0; i < 18; i++) 
    m_fitness_vector[i] = NOT_INITIALIZED;
}

genetic_expr::~genetic_expr ()
/////////////////////////////////////////////
// Deletes the operands.
/////////////////////////////////////////////
{
  for (uint32 i = 0; i < m_num_rands; i++)
    delete m_rands[i];
  delete[] m_rands;
  if (m_pdata) delete m_pdata;
}


void 
finch_claim (bool test, 
	     char *message) 
////////////////////////////////////////////////////////////
//  If the test doesn't hold, this procedure gives an error
//  message and then quits.
////////////////////////////////////////////////////////////
{
  if (!test) 
  {
    cout << "Error: " << message << endl;
    exit (1);
  }
}



