/************************************************************************
 *
 *   File: compile_harness.h:
 *   
 *   Author: Mark Stephenson
 *   Date: Wed Aug 21 19:52:38 2002
 *
 *   Function: Base class for the various harnesses we use.
 *
 ***********************************************************************/

#ifndef __COMPILE_HARNESS_H__
#define __COMPILE_HARNESS_H__


#include "genetic_expr.h"
#include "benchmark_set.h"
#include <vector>
#include <set>
#include <iostream>
#include <functional>
#include <math.h>
#include <bits/nan.h>


class compile_harness
{
 public:

  // Create a new harness.
  compile_harness (uint32  population_size,
		   uint32  generations,
		   double  mutate_rate,
		   double  mortality,
		   char*   initial_pop_fname,
		   char*   save_pop_fname,
		   char*   finch_home,
		   uint32  tournament_size,
		   uint32  dss_set_size,
		   uint32  maximum_genome_size,
		   uint32  max_height,
		   uint32  elitist,
		   uint32  sunset,
		   bool    quantize_results,
		   bool    classifier,
		   benchmark_set *benches);

  
  // Kill the harness.
  virtual ~compile_harness ();

  // Searches for the best expression to the problem.
  virtual genetic_expr *search_for_best ();

 protected:

  // Creates the configuration file.
  void create_configuration ();

  // Creates the initial population file
  void create_initial_population (uint32, uint32, char*);

  // Function that clears whatever directory it'll run in.
  virtual void clear_evaluate_dir () = 0;

  // Function that farms out the jobs and collects results.
  virtual void farm_out_jobs () = 0;

  // Performs crossover and mutation.
  virtual void crossover_and_mutate ();

  // Computes the overall fitness for each expression in 
  // the population.  Also computes optimizability.
  virtual void compute_overall_fitness (uint32 num_benches);
  
  // Prints the population.
  void print_population (ostream&, uint32, bool = false) const;
  
  // Prepares for next generation.
  void prepare_for_generation ();
  
  // Splits fitnesses into quanta.
  void quantize_population ();

  // Returns the population size.
  uint32 popsize () const   { return m_population.size(); }

  // Returns the benchmark set that we're training on.
  benchmark_set *benches () { return m_benches; }

  // Returns the working directory.
  string& workdir ()        { return m_workdir; }
  
  // Returns the evaluation directory.
  string& eval_dir ()       { return m_eval_dir; }

  // Updates the state file with the given fitness.
  void write_state (benchmark_info*, genetic_expr*, uint32);

  // Returns the DSS size to use.
  virtual uint32 dss_size () const;

  // Returns the population.
  vector<genetic_expr*>& population() { return m_population; }

 protected:

  string  m_workdir;
  string  m_eval_dir;
  string  m_run_dir;
  uint32  m_memoized;
  uint32  m_gens;
  double  m_mutate_rate;
  double  m_mortality;
  char*   m_save_pop_fname;

  ofstream *m_state_file;

  uint32  m_tournament_size;
  uint32  m_dss_set_size;
  uint32  m_max_genome_size;
  uint32  m_elitist;
  uint32  m_sunset;

  benchmark_set *m_benches;

  bool m_quantize_results;
  bool m_classifier;
  genetic_expr_bnf *m_bnf;
  set<double> *m_quanta;

  vector<genetic_expr*> m_population;
};


struct expr_lt : public binary_function<genetic_expr*,genetic_expr*,bool>
{
  // Returns true if x is `more fit' than y (i.e., x has a 
  // higher speedup number).
  bool operator () (genetic_expr *x, genetic_expr *y)
  {
    assert (x && y);
    double fity = y->get_fitness ();
    if (isnan(fity)) return true;
    double fitx = x->get_fitness ();
    if (isnan(fitx)) return false;
    assert (fitx > 0.0 && fity > 0.0);

    // Enforce parsimony.
    if (fitx == fity)
      return (x->size() < y->size());
    else
      return (fitx < fity);
  }
};


#endif
